
from typing import cast

from logging import Logger
from logging import getLogger

from abc import ABC
from abc import abstractmethod

from wx import FD_CHANGE_DIR
from wx import FD_OVERWRITE_PROMPT
from wx import FD_SAVE
from wx import FileSelector

from wx import Yield as wxYield

from umlextensions.ExtensionsTypes import FrameInformation
from umlextensions.IExtensionsFacade import IExtensionsFacade

from umlextensions.extensiontypes.BaseExtension import BaseExtension
from umlextensions.extensiontypes.ExtensionDataTypes import UNSPECIFIED_DESCRIPTION
from umlextensions.extensiontypes.ExtensionDataTypes import UNSPECIFIED_FILE_SUFFIX
from umlextensions.extensiontypes.ExtensionDataTypes import UNSPECIFIED_NAME
from umlextensions.extensiontypes.SingleFileRequestResponse import SingleFileRequestResponse

from umlextensions.output.OutputFormat import OutputFormat


class BaseOutputExtension(BaseExtension, ABC):
    """
        Base class for extensions that can convert UML Diagrams into foreign
        structured data.  Examples include but are not limited to:

        * Mermaid
        * PDF
        * Images (png, jpg, bmp, etc
        * Generate Python code
        * Generate Java code

    """
    def __init__(self, extensionsFacade: IExtensionsFacade):
        super().__init__(extensionsFacade)
        self.logger: Logger = getLogger(__name__)

        self._outputFormat:     OutputFormat     = OutputFormat(formatName=UNSPECIFIED_NAME, fileSuffix=UNSPECIFIED_FILE_SUFFIX, description=UNSPECIFIED_DESCRIPTION)
        self._frameInformation: FrameInformation = cast(FrameInformation, None)

    @property
    def outputFormat(self) -> OutputFormat:
        """
        Implementations set the protected variable at class construction

        Returns: The output format type;  Plugins should return `None` if they do
        not support output operations
        """
        return self._outputFormat

    def executeExport(self):
        """
        Called by the extension manager to begin the export process.
        """
        # noinspection PyTypeChecker
        self._extensionsFacade.requestCurrentFrameInformation(callback=self._executeExport)   # type ignore

    def _executeExport(self, frameInformation: FrameInformation):
        """
        The callback necessary to start the export process;

        Args:
            frameInformation:
        """
        assert self._outputFormat.formatName != UNSPECIFIED_NAME,         'Developer error. We cannot export w/o an export format name'
        assert self._outputFormat.fileSuffix != UNSPECIFIED_FILE_SUFFIX,  'Developer error. We cannot export w/o an export file suffix'
        assert self._outputFormat.description != UNSPECIFIED_DESCRIPTION, 'Developer error. We cannot export w/o an export description'

        self._frameInformation = frameInformation

        if self._requireActiveFrame is True:
            if frameInformation.frameActive is False:
                self.showNoUmlFrameDialog()
                return
        if self.setExportOptions() is True:
            self.write()

    def askForFileToExport(self, defaultFileName: str = '', defaultPath: str = '') -> SingleFileRequestResponse:
        """
        Called by a plugin to ask for the export filename

        Returns: The appropriate response object
        """
        wxYield()

        outputFormat: OutputFormat = self.outputFormat
        wildCard:    str = f'{outputFormat.formatName} (*.{outputFormat.fileSuffix}) |*.{outputFormat.fileSuffix}'
        fileName:    str = FileSelector("Choose export file name",
                                        default_path=defaultPath,
                                        default_filename=defaultFileName,
                                        wildcard=wildCard,
                                        flags=FD_SAVE | FD_OVERWRITE_PROMPT | FD_CHANGE_DIR)

        response: SingleFileRequestResponse = SingleFileRequestResponse(cancelled=False)

        if fileName == '':
            response.fileName  = ''
            response.cancelled = True
        else:
            response.fileName = fileName

        return response

    @abstractmethod
    def setExportOptions(self) -> bool:
        """
        Prepare for the export.
        Use this method to query the end-user for any additional export options

        Returns:
            if False, the export is cancelled
        """
        pass

    @abstractmethod
    def write(self) -> bool:
        """
        Write data from a file;  Presumably, the file was specified on the call
        to setExportOptions
        """
        pass




from typing import cast

from logging import Logger
from logging import getLogger

from abc import ABC
from abc import abstractmethod

from wx import DirDialog
from wx import ID_CANCEL
from wx import DD_NEW_DIR_BUTTON

from umlextensions.ExtensionsTypes import FrameInformation
from umlextensions.IExtensionsFacade import IExtensionsFacade
from umlextensions.extensiontypes.BaseExtension import BaseExtension
from umlextensions.extensiontypes.ExtensionDataTypes import UNSPECIFIED_DESCRIPTION
from umlextensions.extensiontypes.ExtensionDataTypes import UNSPECIFIED_FILE_SUFFIX
from umlextensions.extensiontypes.ExtensionDataTypes import UNSPECIFIED_NAME
from umlextensions.input.ImportDirectoryResponse import ImportDirectoryResponse

from umlextensions.input.InputFormat import InputFormat


class BaseInputExtension(BaseExtension, ABC):
    """
    Interface for extensions that can convert foreign structured
    data into UML Diagrams.  Examples include but are not limited to:

        * DTDs
        * Java code
        * Python code

    """
    def __init__(self, extensionsFacade: IExtensionsFacade):

        super().__init__(extensionsFacade)
        self._bieLogger: Logger = getLogger(__name__)

        self._inputFormat:      InputFormat      = InputFormat(formatName=UNSPECIFIED_NAME, fileSuffix=UNSPECIFIED_FILE_SUFFIX, description=UNSPECIFIED_DESCRIPTION)
        self._frameInformation: FrameInformation = cast(FrameInformation, None)

    @property
    def inputFormat(self) -> InputFormat:
        """
        Implementations set the protected variable at class construction

        Returns: The input format type; Plugins should return `None` if they do
        not support input operations
        """
        return self._inputFormat

    def askForImportDirectoryName(self) -> ImportDirectoryResponse:
        """
        Called by extension to ask which directory must be imported

        Returns:  The appropriate response object;  The directory name is valid only if
        response.cancelled is True
        """
        # defaultPath: str = self._pluginAdapter.currentDirectory
        defaultPath: str = ''
        dirDialog: DirDialog = DirDialog(None,
                                         "Choose a directory to import",
                                         defaultPath=defaultPath,
                                         style=DD_NEW_DIR_BUTTON)

        response: ImportDirectoryResponse = ImportDirectoryResponse()
        if dirDialog.ShowModal() == ID_CANCEL:
            response.cancelled     = True
            response.directoryName = ''
        else:
            response.directoryName = dirDialog.GetPath()
            response.cancelled     = False
            # self._pluginAdapter.currentDirectory = response.directoryName    # TODO: Should plugin be doing this?  No

        dirDialog.Destroy()

        return response

    def executeImport(self):
        """
        Called by the extension manager to begin the import process.
        """
        # noinspection PyTypeChecker
        self._extensionsFacade.requestCurrentFrameInformation(callback=self._executeImport)   # type ignore

    def _executeImport(self, frameInformation: FrameInformation):
        """
        The callback necessary to start the import process;
        Args:
            frameInformation:
        """
        assert self._inputFormat.formatName != UNSPECIFIED_NAME,         'Developer error. We cannot export w/o an import format name'
        assert self._inputFormat.fileSuffix != UNSPECIFIED_FILE_SUFFIX,  'Developer error. We cannot export w/o an import file suffix'
        assert self._inputFormat.description != UNSPECIFIED_DESCRIPTION, 'Developer error. We cannot import w/o an import description'

        self._frameInformation = frameInformation

        if self._requireActiveFrame is True:
            if frameInformation.frameActive is False:
                self.showNoUmlFrameDialog()
                return
        if self.setImportOptions() is True:
            self.read()

    @abstractmethod
    def setImportOptions(self) -> bool:
        """
        Prepare for the import.
        Use this method to query the end-user for any additional import options

        Returns:
            if False, the import is cancelled
        """
        pass

    @abstractmethod
    def read(self) -> bool:
        """
        Read data from a file;  Presumably, the file was specified on the call
        to setImportOptions
        """
        pass

    def _composeWildCardSpecification(self) -> str:

        inputFormat: InputFormat = self.inputFormat

        # wildcard: str = inputFormat.name + " (*." + inputFormat.extension + ")|*." + inputFormat.description
        wildcard: str = (
            f'{inputFormat.formatName} '
            f' (*, {inputFormat.fileSuffix}) '
            f'|*.{inputFormat.fileSuffix}'
        )
        return wildcard



from typing import cast

from logging import Logger
from logging import getLogger

from wx import OK
from wx import MessageBox
from wx import ICON_ERROR

from umlshapes.ShapeTypes import UmlShapes

from umlshapes.shapes.UmlNote import UmlNote
from umlshapes.shapes.UmlClass import UmlClass

from umlextensions.IExtensionsFacade import IExtensionsFacade

from umlshapes.mixins.TopLeftMixin import TopLeftMixin

from umlshapes.types.UmlPosition import UmlPosition

from umlextensions.extensiontypes.ExtensionDataTypes import Author
from umlextensions.extensiontypes.ExtensionDataTypes import Version
from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionName

from umlextensions.tools.orthogonallayout.OrthogonalAdapter import OrthogonalAdapter
from umlextensions.tools.orthogonallayout.OrthogonalAdapter import UmlShapeCoordinate
from umlextensions.tools.orthogonallayout.OrthogonalAdapter import UmlShapeCoordinates
from umlextensions.tools.orthogonallayout.OrthogonalAdapterException import OrthogonalAdapterException

from umlextensions.tools.orthogonallayout.DlgLayoutDimensions import DlgLayoutDimensions
from umlextensions.tools.orthogonallayout.LayoutAreaDimensions import LayoutAreaDimensions

from umlextensions.tools.BaseToolExtension import BaseToolExtension


class ToolOrthogonalLayout(BaseToolExtension):

    def __init__(self, extensionsFacade: IExtensionsFacade):

        super().__init__(extensionsFacade=extensionsFacade)

        self.logger: Logger = getLogger(__name__)

        self._name      = ExtensionName('Orthogonal Layout')
        self._author    = Author('Humberto A. Sanchez II')
        self._version   = Version('3.0')

        self._layoutWidth:  int = 0
        self._layoutHeight: int = 0

    def setOptions(self) -> bool:
        with DlgLayoutDimensions(None) as dlg:    # noqa
            if dlg.ShowModal() == OK:
                self.logger.warning(f'Retrieved data: layoutWidth: {dlg.layoutWidth} layoutHeight: {dlg.layoutHeight}')
                self._layoutWidth  = dlg.layoutWidth
                self._layoutHeight = dlg.layoutHeight
                proceed: bool = True
            else:
                self.logger.info(f'Cancelled')
                proceed = False

        return proceed

    def doAction(self):
        self.logger.info(f'Begin Orthogonal Routing')
        self._extensionsFacade.getSelectedUmlShapes(callback=self._doAction)
        self.logger.info('End Orthogonal Routing')

    def _doAction(self, selectedUmlShapes):

        try:
            orthogonalAdapter: OrthogonalAdapter = OrthogonalAdapter(umlShapes=selectedUmlShapes)

            layoutAreaSize: LayoutAreaDimensions = LayoutAreaDimensions(self._layoutWidth, self._layoutHeight)
            orthogonalAdapter.doLayout(layoutAreaSize)
        except OrthogonalAdapterException as oae:
            MessageBox(f'{oae}', 'Error', OK | ICON_ERROR)
            return

        if orthogonalAdapter is not None:
            self._reLayoutNodes(selectedUmlShapes, orthogonalAdapter.umlShapeCoordinates)

            self._extensionsFacade.wiggleShapes()
            self._extensionsFacade.extensionModifiedProject()

    def _reLayoutNodes(self, umlObjects: UmlShapes, umlShapeCoordinates: UmlShapeCoordinates):
        """

        Args:
            umlObjects:
        """
        handledShapes = {UmlClass, UmlNote}
        for umlObj in umlObjects:
            if type(umlObj) in handledShapes:
                if isinstance(umlObj, UmlClass):
                    umlName: str = umlObj.modelClass.name
                else:
                    umlName = cast(UmlNote, umlObj).modelNote.name

                umlShapeCoordinate: UmlShapeCoordinate = umlShapeCoordinates[umlName]

                self._moveShape(cast(TopLeftMixin, umlObj), umlShapeCoordinate)      # noqas

    def _moveShape(self, srcShape: TopLeftMixin, umlCoordinate: UmlShapeCoordinate):
        """

        Args:
            srcShape:
            umlCoordinate:
        """

        oldPosition: UmlPosition = srcShape.position

        umlPosition: UmlPosition = self._preferences.orthogonalLayoutTopLeft
        newPosition: UmlPosition = UmlPosition(x=umlCoordinate.x + umlPosition.x, y=umlCoordinate.y + umlPosition.y)
        self.logger.info(f'{srcShape} - {oldPosition=}  {newPosition=}')

        srcShape.position = newPosition




from typing import List
from typing import NewType
from typing import cast

from logging import Logger
from logging import getLogger

from dataclasses import dataclass

from wx import BORDER_DEFAULT
from wx import CANCEL
from wx import CommandEvent
from wx import DEFAULT_DIALOG_STYLE
from wx import EVT_BUTTON
from wx import EVT_CLOSE
from wx import EVT_SPINCTRL
from wx import EVT_SPINCTRLDOUBLE
from wx import ID_ANY
from wx import ID_CANCEL
from wx import ID_OK
from wx import OK
from wx import RESIZE_BORDER
from wx import Size
from wx import SpinCtrl
from wx import SpinCtrlDouble
from wx import SpinDoubleEvent
from wx import StaticText
from wx import StdDialogButtonSizer
from wx import Window

from wx import NewIdRef as wxNewIdRef

from wx.lib.sized_controls import SizedDialog
from wx.lib.sized_controls import SizedPanel
from wx.lib.sized_controls import SizedStaticBox

from pyorthogonalrouting.Configuration import Configuration
from pyorthogonalrouting.Rect import Rect

from umlextensions.ExtensionsTypes import ShapeBoundaries
from umlextensions.IExtensionsFacade import IExtensionsFacade
from umlextensions.tools.orthogonalrouting.LabelledSlider import LabelledSlider


@dataclass
class GlobalBoundsControl:
    label:        str
    spinCtrl:     SpinCtrl
    value:        int
    minValue:     int
    maxValue:     int
    id:           int


MIN_GLOBAL_BOUND: int = 0
MAX_GLOBAL_BOUND: int = 10000

GlobalBoundsControls = NewType('GlobalBoundsControls', List[GlobalBoundsControl])

NO_SPIN_CTRL: SpinCtrl = cast(SpinCtrl, None)


class DlgOrthogonalRoutingConfiguration(SizedDialog):

    def __init__(self, parent: Window, extensionsFacade: IExtensionsFacade):
        """
        
        Args:
            parent: 
            extensionsFacade: 
        """

        self._extensionsFacade: IExtensionsFacade = extensionsFacade
        self._configuration: Configuration     = cast(Configuration, None)

        style:   int  = DEFAULT_DIALOG_STYLE | RESIZE_BORDER
        dlgSize: Size = Size(475, 400)

        super().__init__(parent, title='Configuration for Orthogonal Connector Routing', size=dlgSize, style=style)
        self.logger: Logger = getLogger(__name__)

        self._leftId:   int = wxNewIdRef()
        self._topId:    int = wxNewIdRef()
        self._widthId:  int = wxNewIdRef()
        self._heightId: int = wxNewIdRef()

        self._left:   SpinCtrl = NO_SPIN_CTRL
        self._top:    SpinCtrl = NO_SPIN_CTRL
        self._width:  SpinCtrl = NO_SPIN_CTRL
        self._height: SpinCtrl = NO_SPIN_CTRL

        self._sizedPanel: SizedPanel = self.GetContentsPane()
        self._sizedPanel.SetSizerType('horizontal')

        self._sourceEdgeDistance:      SpinCtrlDouble = cast(SpinCtrlDouble, None)
        self._destinationEdgeDistance: SpinCtrlDouble = cast(SpinCtrlDouble, None)

        self._extensionsFacade.getShapeBoundaries(callback=self._objectBoundariesCallback)
        #
        # Doing this since that it appears that using the above callback messes
        # with the layout
        #
        self.PostSizeEvent()

    def _objectBoundariesCallback(self, objectBoundaries: ShapeBoundaries):
        """
        Get the updated object boundaries

        Args:
            objectBoundaries:

        """
        self._configuration = Configuration()

        width:     int  = objectBoundaries.maxX - objectBoundaries.minX
        height:    int  = objectBoundaries.maxY - objectBoundaries.minY
        newBounds: Rect = Rect(top=objectBoundaries.minY, left=objectBoundaries.minX, width=width, height=height)

        self._configuration.globalBounds = newBounds

        self.logger.info(f'{self._configuration.globalBounds=}')
        self._doLayout()

    def _doLayout(self):

        self._layoutControls(parent=self._sizedPanel)
        self._layoutStandardOkCancelButtonSizer()

    def _layoutStandardOkCancelButtonSizer(self):
        """
        Call this last when creating controls; Will take care of
        adding callbacks for the Ok and Cancel buttons
        """
        buttSizer: StdDialogButtonSizer = self.CreateStdDialogButtonSizer(OK | CANCEL)

        self.SetButtonSizer(buttSizer)
        self.Bind(EVT_BUTTON, self._onOk,    id=ID_OK)
        self.Bind(EVT_BUTTON, self._onClose, id=ID_CANCEL)
        self.Bind(EVT_CLOSE,  self._onClose)

    def _layoutControls(self, parent: SizedPanel):

        localPanel: SizedPanel = SizedPanel(parent, style=BORDER_DEFAULT)
        localPanel.SetSizerType('vertical')
        localPanel.SetSizerProps(expand=True, proportion=2)

        shapeMargin:        LabelledSlider = LabelledSlider(sizedPanel=localPanel, label='Shape Margin',
                                                            value=self._configuration.shapeMargin,
                                                            minValue=0,
                                                            maxValue=100,
                                                            size=Size(325, height=-1),
                                                            toolTip='The margin around shapes for routing')
        globalBoundsMargin: LabelledSlider = LabelledSlider(sizedPanel=localPanel, label='Global Bounds Margin',
                                                            value=self._configuration.globalBoundsMargin,
                                                            minValue=0,
                                                            maxValue=100,
                                                            size=Size(325, height=-1),
                                                            toolTip='The margin that routing expands')

        shapeMargin.valueChangedHandler        = self._shapeMarginChanged
        globalBoundsMargin.valueChangedHandler = self._globalBoundsMarginChanged

        self._layoutConnectorEdgeDistance(parent=localPanel)
        self._layoutGlobalBounds(parent=parent)

    def _layoutConnectorEdgeDistance(self, parent: SizedPanel):

        sourceBox: SizedStaticBox = SizedStaticBox(parent=parent, label='Source connector distance')
        sourceBox.SetSizerType('vertical')
        sourceBox.SetSizerProps(proportion=1)

        sourceEdgeDistance: SpinCtrlDouble = SpinCtrlDouble(sourceBox, id=ID_ANY,
                                                            value=str(self._configuration.sourceEdgeDistance),
                                                            min=0.0,
                                                            max=1.0,
                                                            inc=0.1
                                                            )
        sourceEdgeDistance.SetToolTip('Ratio of where to place connectors on source shape edge.')

        destinationBox: SizedStaticBox = SizedStaticBox(parent=parent, label='Destination connector distance')
        destinationBox.SetSizerType('vertical')
        destinationBox.SetSizerProps(proportion=1)

        destinationEdgeDistance: SpinCtrlDouble = SpinCtrlDouble(destinationBox, id=ID_ANY,
                                                                 value=str(self._configuration.destinationEdgeDistance),
                                                                 min=0.0,
                                                                 max=1.0,
                                                                 inc=0.1
                                                                 )
        destinationEdgeDistance.SetToolTip('Ratio of where to place connectors on destination shape edge.')

        parent.Bind(EVT_SPINCTRLDOUBLE, handler=self._onSourceEdgeDistancedChanged,      source=sourceEdgeDistance)
        parent.Bind(EVT_SPINCTRLDOUBLE, handler=self._onDestinationEdgeDistancedChanged, source=destinationEdgeDistance)

    def _layoutGlobalBounds(self, parent: SizedPanel):

        configuration: Configuration = self._configuration

        globalBoundsControls: GlobalBoundsControls = GlobalBoundsControls(
            [
                GlobalBoundsControl('Left:',    self._left,   configuration.globalBounds.left,   MIN_GLOBAL_BOUND, MAX_GLOBAL_BOUND, id=self._leftId),
                GlobalBoundsControl('Top: ',    self._top,    configuration.globalBounds.top,    MIN_GLOBAL_BOUND, MAX_GLOBAL_BOUND, id=self._topId),
                GlobalBoundsControl('Width:',   self._width,  configuration.globalBounds.width,  MIN_GLOBAL_BOUND, MAX_GLOBAL_BOUND, id=self._widthId),
                GlobalBoundsControl('Height: ', self._height, configuration.globalBounds.height, MIN_GLOBAL_BOUND, MAX_GLOBAL_BOUND, id=self._heightId),
            ]
        )

        labelBox: SizedStaticBox = SizedStaticBox(parent=parent, label='Global Bounds')
        labelBox.SetSizerProps(expand=True, proportion=1)
        labelBox.SetToolTip('Defines the routing confinement bounds')

        localPanel: SizedPanel = SizedPanel(labelBox)
        localPanel.SetSizerType('form')
        localPanel.SetSizerProps(expand=True)

        for c in globalBoundsControls:
            control: GlobalBoundsControl = cast(GlobalBoundsControl, c)
            StaticText(parent=localPanel, label=control.label)

            control.spinCtrl = SpinCtrl(localPanel, id=control.id, size=Size(25, -1))
            control.spinCtrl.SetRange(MIN_GLOBAL_BOUND, MAX_GLOBAL_BOUND)
            control.spinCtrl.SetValue(control.value)
            control.spinCtrl.SetSizerProps(expand=True)

            self.Bind(EVT_SPINCTRL, handler=self._onGlobalBoundChanged, source=control.spinCtrl)

    # noinspection PyUnusedLocal
    def _onOk(self, event: CommandEvent):
        """
        """
        self.EndModal(OK)

    # noinspection PyUnusedLocal
    def _onClose(self, event: CommandEvent):
        """
        """
        self.EndModal(CANCEL)

    def _shapeMarginChanged(self, event: CommandEvent):
        self._configuration.shapeMargin = event.GetInt()

    def _globalBoundsMarginChanged(self, event: CommandEvent):
        self._configuration.globalBoundsMargin = event.GetInt()

    def _onGlobalBoundChanged(self, event: CommandEvent):

        bounds:   Rect = self._configuration.globalBounds
        newValue: int  = event.GetInt()
        match event.GetId():
            case self._leftId:
                bounds.left = newValue
            case self._topId:
                bounds.top = newValue
            case self._widthId:
                bounds.width = newValue
            case self._heightId:
                bounds.height = newValue
            case _:
                self.logger.error(f'Unknown control id')

        self._configuration.globalBounds = bounds

    def _onSourceEdgeDistancedChanged(self, event: SpinDoubleEvent):

        newValue: float = event.GetValue()
        self._configuration.sourceEdgeDistance = newValue

    def _onDestinationEdgeDistancedChanged(self, event: SpinDoubleEvent):

        newValue: float = event.GetValue()
        self._configuration.destinationEdgeDistance = newValue




from typing import cast

from logging import Logger
from logging import getLogger

from wx import CANCEL
from wx import DEFAULT_DIALOG_STYLE
from wx import EVT_BUTTON
from wx import EVT_CHECKBOX
from wx import EVT_CLOSE
from wx import FONTFAMILY_SWISS
from wx import FONTSTYLE_NORMAL
from wx import FONTWEIGHT_BOLD
from wx import ID_CANCEL
from wx import ID_OK
from wx import OK
from wx import RESIZE_BORDER

from wx import Font
from wx import CheckBox
from wx import CommandEvent
from wx import STAY_ON_TOP
from wx import StaticText
from wx import StdDialogButtonSizer
from wx import Window

from wx.lib.sized_controls import SizedDialog
from wx.lib.sized_controls import SizedPanel

from umlextensions.ExtensionsTypes import DiagnosticInformation
from umlextensions.ExtensionsTypes import IntegerList
from umlextensions.ExtensionsTypes import Points
from umlextensions.ExtensionsTypes import Rectangle
from umlextensions.ExtensionsTypes import Rectangles
from umlextensions.IExtensionsFacade import IExtensionsFacade


class DlgDiagnoseLayout(SizedDialog):

    def __init__(self, parent: Window | None):

        style:   int  = DEFAULT_DIALOG_STYLE | RESIZE_BORDER | STAY_ON_TOP

        super().__init__(parent, title='Orthogonal Connector Routing Diagnosis', style=style)
        self.logger: Logger = getLogger(__name__)

        self._showRulers:          CheckBox = cast(CheckBox, None)
        self._showReferencePoints: CheckBox = cast(CheckBox, None)
        self._showRouteGrid:       CheckBox = cast(CheckBox, None)

        self._diagnosticInformation: DiagnosticInformation = cast(DiagnosticInformation, None)
        self._extensionsFacade:      IExtensionsFacade     = cast(IExtensionsFacade, None)

        self._doLayout(self.GetContentsPane())
        self._layoutStandardOkButtonSizer()

        self._bindCallbacks(parent=self)

        # self.Fit()
        # self.SetMinSize(self.GetSize())

    def _setDiagnosticInformation(self, diagnosticInformation: DiagnosticInformation):
        self._diagnosticInformation = diagnosticInformation

    def _setExtensionsFacade(self, pluginAdapter: IExtensionsFacade):
        self._pluginAdapter = pluginAdapter

    # noinspection PyTypeChecker
    diagnosticInformation = property(fget=None, fset=_setDiagnosticInformation, doc='Write only diagnostic information')
    # noinspection PyTypeChecker
    extensionsFacade         = property(fget=None, fset=_setExtensionsFacade, doc='Extensions Facade to talk to application')

    def _doLayout(self, parentPanel: SizedPanel):
        self._layoutAlgorithmLayers(sizedPanel=parentPanel)

    def _layoutAlgorithmLayers(self, sizedPanel: SizedPanel):
        container: SizedPanel = SizedPanel(sizedPanel)
        container.SetSizerType('vertical')
        # (["top", "left", "right"], 6))
        container.SetSizerProps(expand=True, proprtion=1, border=(["top"], 24))

        font:  Font       = Font(18, FONTFAMILY_SWISS, FONTSTYLE_NORMAL, FONTWEIGHT_BOLD)
        title: StaticText = StaticText(container, label='Algorithm Layers')
        title.SetFont(font)

        self._showRulers          = CheckBox(container, label='Rulers')
        self._showReferencePoints = CheckBox(container, label='Reference Points')
        self._showRouteGrid       = CheckBox(container, label='Route Grid')

    def _layoutStandardOkButtonSizer(self):
        """
        Call this last when creating controls; Will take care of
        adding callbacks for the Ok and Cancel buttons
        """
        buttSizer: StdDialogButtonSizer = self.CreateStdDialogButtonSizer(OK)

        self.SetButtonSizer(buttSizer)
        self.Bind(EVT_BUTTON, self._onOk,    id=ID_OK)
        self.Bind(EVT_BUTTON, self._onClose, id=ID_CANCEL)
        self.Bind(EVT_CLOSE,  self._onClose)

    def _bindCallbacks(self, parent):

        parent.Bind(EVT_CHECKBOX, self._onShowRulers,          self._showRulers)
        parent.Bind(EVT_CHECKBOX, self._onShowReferencePoints, self._showReferencePoints)
        parent.Bind(EVT_CHECKBOX, self._onShowRouteGrid,       self._showRouteGrid)

    # noinspection PyUnusedLocal
    def _onOk(self, event: CommandEvent):
        """
        Typically shown non modal
        """
        if self.IsModal() is True:
            self.EndModal(OK)
        else:
            self.Destroy()

    # noinspection PyUnusedLocal
    def _onClose(self, event: CommandEvent):
        """
        Typically shown non modal
1        """
        if self.IsModal() is True:
            self.EndModal(CANCEL)
        else:
            self.Destroy()

    def _onShowRulers(self, event: CommandEvent):

        self._ensureSetupOk()

        if event.IsChecked() is True:
            self._pluginAdapter.showRulers(show=True,
                                           horizontalRulers=self._diagnosticInformation.horizontalRulers,
                                           verticalRulers=self._diagnosticInformation.verticalRulers,
                                           diagramBounds=self._diagnosticInformation.diagramBounds)
        else:
            self._pluginAdapter.showRulers(show=False,
                                           verticalRulers=cast(IntegerList, None),
                                           horizontalRulers=cast(IntegerList, None),
                                           diagramBounds=cast(Rectangle, None)
                                           )

    def _onShowReferencePoints(self, event: CommandEvent):

        self._ensureSetupOk()

        if event.IsChecked() is True:
            self._pluginAdapter.showOrthogonalRoutingPoints(show=True, spots=self._diagnosticInformation.spots)
        else:
            self._pluginAdapter.showOrthogonalRoutingPoints(show=False, spots=cast(Points, None))

    def _onShowRouteGrid(self, event: CommandEvent):

        self._ensureSetupOk()
        if event.IsChecked() is True:
            self._pluginAdapter.showRouteGrid(show=True, routeGrid=self._diagnosticInformation.routeGrid)
        else:
            self._pluginAdapter.showRouteGrid(show=False, routeGrid=cast(Rectangles, None))

    def _ensureSetupOk(self):
        """
        I know, I know this does nothing when assertions are off;  But this is
        a developer demo and they will be on
        """
        assert self._pluginAdapter is not None, 'Developer error.  Forgot to inject the manager'
        assert self._diagnosticInformation is not None, 'Developer error.  Forgot to inject diagnostic information'




from typing import Set
from typing import cast

from logging import Logger
from logging import getLogger

from os import linesep as osLineSep

from umlmodel.Class import Class
from umlmodel.Note import Note

from umlshapes.ShapeTypes import UmlShapeGenre
from umlshapes.ShapeTypes import UmlShapes

from umlshapes.links.UmlLink import UmlLink
from umlshapes.mixins.IdentifierMixin import IdentifierMixin

from umlshapes.shapes.UmlClass import UmlClass
from umlshapes.shapes.UmlNote import UmlNote
from umlshapes.types.Common import EndPositions

from umlshapes.types.UmlDimensions import UmlDimensions
from umlshapes.types.UmlPosition import UmlPosition
from umlshapes.types.UmlPosition import UmlPositions

from umlextensions.output.gml.UnsupportedOperation import UnsupportedOperation

ExportableShapeType = UmlClass | UmlNote
ExportableModelType = Class | Note


class GMLExporter:

    VERSION: str = '2.0.0'

    GRAPH_TOKEN: str = 'graph'

    ID_TOKEN:    str = 'id'
    LABEL_TOKEN: str = 'label'
    NODE_TOKEN:  str = 'node'
    EDGE_TOKEN:  str = 'edge'

    SOURCE_ID_TOKEN: str = 'source'
    TARGET_ID_TOKEN: str = 'target'

    GRAPHICS_TOKEN: str = 'graphics'
    START_TOKEN:    str = '['
    END_TOKEN:      str = ']'

    QUOTE_TOKEN: str = '"'

    LINE_DEFINITION_TOKEN:  str = 'Line'
    POINT_DEFINITION_TOKEN: str = 'point'

    X_POSITION_TOKEN: str = 'x'
    Y_POSITION_TOKEN: str = 'y'
    Z_POSITION_TOKEN: str = 'z'
    WIDTH_TOKEN:      str = 'w'
    HEIGHT_TOKEN:     str = 'h'
    DEPTH_TOKEN:      str = 'd'

    singleTab: str = ''
    doubleTab: str = ''
    tripleTab: str = ''
    quadrupleTab: str = ''
    quintupleTab: str = ''

    def __init__(self):

        self.logger:   Logger = getLogger(__name__)
        self._gml:     str    = ''

        self._prettyPrint: bool = True

    def translate(self, umlShapes: UmlShapes):

        if self._prettyPrint is True:
            GMLExporter.singleTab = '\t'
            GMLExporter.doubleTab = '\t\t'
            GMLExporter.tripleTab = '\t\t\t'
            GMLExporter.quadrupleTab = '\t\t\t\t'
            GMLExporter.quintupleTab = '\t\t\t\t\t'

        gml: str = self._generateGraphStart()
        gml = self._generateNodes(umlShapes, gml)
        gml = self._generateUniqueEdges(umlShapes=umlShapes, gml=gml)

        gml = self._generateGraphTermination(gml)
        self._gml = gml

    @property
    def gml(self):
        return self._gml

    @gml.setter
    def gml(self, theNewValue):
        raise UnsupportedOperation('gml is a read-only property')

    @property
    def prettyPrint(self):
        return self._prettyPrint

    @prettyPrint.setter
    def prettyPrint(self, theNewValue):
        self._prettyPrint = theNewValue

    def write(self, pathToFile: str):

        with open(pathToFile, 'w') as writer:
            writer.write(self._gml)

    def _generateNodes(self, umlShapes: UmlShapes, gml: str) -> str:

        nodeGml: str = ''
        for umlShape in umlShapes:
            if isinstance(umlShape, UmlClass) or isinstance(umlShape, UmlNote):

                identifier:  IdentifierMixin     = cast(IdentifierMixin, umlShape)
                if isinstance(umlShape, UmlClass):
                    modelObject: ExportableModelType = umlShape.modelClass
                elif isinstance(umlShape, UmlNote):
                    modelObject = umlShape.modelNote
                else:
                    assert False, 'Unsupported node type'
                nodeGml = (
                    f'{nodeGml}'
                    f'{GMLExporter.singleTab}{GMLExporter.NODE_TOKEN} {GMLExporter.START_TOKEN}\n'
                    f'{GMLExporter.doubleTab}{GMLExporter.ID_TOKEN} "{identifier.id}"\n'
                    f'{GMLExporter.doubleTab}{GMLExporter.LABEL_TOKEN} "{modelObject.name}"\n'
                    f'{self._generateNodeGraphicsSection(umlShape)}'
                    f'{GMLExporter.singleTab}{GMLExporter.END_TOKEN}\n'
                )
                self.logger.debug(f'{nodeGml}')
        return f'{gml}{nodeGml}'

    def _generateNodeGraphicsSection(self, umlShape: UmlShapeGenre) -> str:

        # pos = umlShape.GetPosition()
        position: UmlPosition = umlShape.position
        x = position.x
        y = position.y
        z = 0
        dimensions: UmlDimensions = umlShape.size
        w = dimensions.width
        h = dimensions.height
        d = 0
        gml = (
            f'{GMLExporter.doubleTab}{GMLExporter.GRAPHICS_TOKEN} {GMLExporter.START_TOKEN}\n'
            
            f'{GMLExporter.tripleTab}{GMLExporter.X_POSITION_TOKEN} {x}\n'
            f'{GMLExporter.tripleTab}{GMLExporter.Y_POSITION_TOKEN} {y}\n'
            f'{GMLExporter.tripleTab}{GMLExporter.Z_POSITION_TOKEN} {z}\n'
            f'{GMLExporter.tripleTab}{GMLExporter.WIDTH_TOKEN} {w}\n'
            f'{GMLExporter.tripleTab}{GMLExporter.HEIGHT_TOKEN} {h}\n'
            f'{GMLExporter.tripleTab}{GMLExporter.DEPTH_TOKEN} {d}\n'
            f'{GMLExporter.tripleTab}type "rectangle"\n'
            f'{GMLExporter.tripleTab}width 0.12\n'
            f'{GMLExporter.tripleTab}fill "#ff0000"\n'
            f'{GMLExporter.tripleTab}outline "#000000"\n'
            
            f'{GMLExporter.doubleTab}{GMLExporter.END_TOKEN}\n'
        )
        return gml

    def _generateGraphStart(self, graphName: str = 'DefaultGraphName') -> str:

        gml: str = (
            f'{GMLExporter.GRAPH_TOKEN} {GMLExporter.START_TOKEN}\n'
            f'{GMLExporter.singleTab}directed 1\n'
            f'{GMLExporter.singleTab}version  1.0\n'
            f'{GMLExporter.singleTab}label "GML for {graphName}"\n'
            f'{GMLExporter.singleTab}comment "Generated by GML Exporter Version {GMLExporter.VERSION}"\n'
        )

        return gml

    def _generateGraphTermination(self, gml: str) -> str:

        gml = f'{gml}{GMLExporter.END_TOKEN}{osLineSep}'
        return gml

    def _generateUniqueEdges(self, umlShapes: UmlShapes, gml: str) -> str:
        from umlshapes.ShapeTypes import UmlLinks

        linkSet: Set = set()        # Concatenated str link ids;  e.g, 1-2

        for umlShape in umlShapes:
            if isinstance(umlShape, UmlClass) or isinstance(umlShape, UmlNote):
                links: UmlLinks = umlShape.links
                self.logger.info(f'links: {links}')
                for umlLink in links:
                    srcShapeId:  str = umlLink.sourceShape.id
                    destShapeId: str = umlLink.destinationShape.id
                    linkIds:   str = f'{srcShapeId}-{destShapeId}'
                    if linkIds not in linkSet:
                        gml = self.__generateUniqueEdge(umlLink=umlLink, gml=gml)
                        linkSet.add(linkIds)

        return gml

    def __generateUniqueEdge(self, umlLink: UmlLink, gml: str) -> str:

        srcShapeId:  str = umlLink.sourceShape.id
        destShapeId: str = umlLink.destinationShape.id

        gml = (
            f'{gml}'
            f'{GMLExporter.singleTab}{GMLExporter.EDGE_TOKEN} {GMLExporter.START_TOKEN}\n'
            f'{GMLExporter.doubleTab}{GMLExporter.ID_TOKEN} "{umlLink.id}"\n'
            f'{GMLExporter.doubleTab}{GMLExporter.SOURCE_ID_TOKEN} "{srcShapeId}"\n'
            f'{GMLExporter.doubleTab}{GMLExporter.TARGET_ID_TOKEN} "{destShapeId}"\n'
            f'{self.__generateEdgeGraphicsSection(umlLink=umlLink)}'
            f'{GMLExporter.singleTab}{GMLExporter.END_TOKEN}\n'
        )

        return gml

    def __generateEdgeGraphicsSection(self, umlLink: UmlLink) -> str:

        endPositions:     EndPositions = umlLink.endPositions
        controlPositions: UmlPositions = umlLink.controlPositions

        edgeGml: str = (
            f'{GMLExporter.doubleTab}{GMLExporter.GRAPHICS_TOKEN} {GMLExporter.START_TOKEN}\n'
            f'{GMLExporter.tripleTab}type "line"\n'
            f'{GMLExporter.tripleTab}arrow "last"\n'
            f'{GMLExporter.tripleTab}{GMLExporter.LINE_DEFINITION_TOKEN} {GMLExporter.START_TOKEN}\n'
            f'{self.__generatePoint(endPositions.fromPosition)}'
            f'{self.__generatePoints(controlPositions)}'
            f'{self.__generatePoint(endPositions.toPosition)}'
            f'{GMLExporter.tripleTab}{GMLExporter.END_TOKEN}\n'
            f'{GMLExporter.doubleTab}{GMLExporter.END_TOKEN}\n'
        )

        return edgeGml

    def __generatePoints(self, controlPoints) -> str:

        pointsGml: str = ''
        for umlPosition in controlPoints:
            pointsGml = (
                f'{pointsGml}{self.__generatePoint(umlPosition)}'
            )

        return pointsGml

    def __generatePoint(self, umlPosition: UmlPosition) -> str:

        # position: Tuple[int, int] = linePoint.GetPosition()

        x:        int = umlPosition.x
        y:        int = umlPosition.y
        z:        int = 0
        pointGml: str = (
            f'{GMLExporter.quadrupleTab}{GMLExporter.POINT_DEFINITION_TOKEN} {GMLExporter.START_TOKEN}\n'
            f'{GMLExporter.quintupleTab}{GMLExporter.X_POSITION_TOKEN} {x}\n'
            f'{GMLExporter.quintupleTab}{GMLExporter.Y_POSITION_TOKEN} {y}\n'
            f'{GMLExporter.quintupleTab}{GMLExporter.Z_POSITION_TOKEN} {z}\n'
            f'{GMLExporter.quadrupleTab}{GMLExporter.END_TOKEN}\n'
        )

        return pointGml




from typing import List

from dataclasses import dataclass


@dataclass
class LayoutAreaDimensions:

    width:  int = 0
    height: int = 0

    @classmethod
    def deSerialize(cls, value: str) -> 'LayoutAreaDimensions':

        layoutAreaSize: LayoutAreaDimensions = LayoutAreaDimensions()

        widthHeight: List[str] = value.split(sep=',')

        if len(widthHeight) != 2:
            raise ValueError(f'Incorrectly formatted dimensions: {value}')

        if not value.replace(',', '', 1).isdigit():
            raise ValueError(f'String must be numeric: {value}')

        layoutAreaSize.width  = int(widthHeight[0])
        layoutAreaSize.height = int(widthHeight[1])

        return layoutAreaSize

    def __str__(self):
        return f'{self.width},{self.height}'

    def __repr__(self):
        return self.__str__()




from typing import Tuple
from typing import cast

from logging import Logger
from logging import getLogger

from codeallybasic.Position import Position

from umlmodel.enumerations.LinkType import LinkType

from umlshapes.ShapeTypes import LinkableUmlShape
from umlshapes.ShapeTypes import UmlLinkGenre
from umlshapes.ShapeTypes import UmlShapeGenre

from umlshapes.links.UmlInheritance import UmlInheritance

from umlshapes.types.Common import EndPositions
from umlshapes.types.UmlDimensions import UmlDimensions

from umlshapes.types.UmlPosition import UmlPosition
from umlshapes.types.UmlPosition import UmlPositions

from pyorthogonalrouting.Common import Integers
from pyorthogonalrouting.OrthogonalConnectorByProduct import OrthogonalConnectorByProduct

from pyorthogonalrouting.Point import Point
from pyorthogonalrouting.Point import Points
from pyorthogonalrouting.Rectangle import Rectangle
from pyorthogonalrouting.Rect import Rect
from pyorthogonalrouting.Configuration import Configuration
from pyorthogonalrouting.ConnectorPoint import ConnectorPoint
from pyorthogonalrouting.OrthogonalConnector import OrthogonalConnector
from pyorthogonalrouting.OrthogonalConnectorOptions import OrthogonalConnectorOptions

from pyorthogonalrouting.enumerations.Side import Side

from umlextensions.ExtensionsTypes import AssociationName
from umlextensions.ExtensionsTypes import DestinationCardinality
from umlextensions.ExtensionsTypes import DiagnosticInformation
from umlextensions.ExtensionsTypes import InterfaceName
from umlextensions.ExtensionsTypes import LinkInformation
from umlextensions.ExtensionsTypes import SourceCardinality
from umlextensions.IExtensionsFacade import IExtensionsFacade

ANCHOR_POINT_ADJUSTMENT: int = 1    # because line end needs to look like it is right on the line


class OrthogonalConnectorAdapter:
    """
    TODO:  This adapter leaks some of the OrthogonalConnector data types.  Fix this before
    merging to the master branch
    Create a single property call diagnosticInformation

    """
    def __init__(self, extensionsFacade: IExtensionsFacade):

        self.logger: Logger = getLogger(__name__)

        self._extensionsFacade: IExtensionsFacade = extensionsFacade
        self._configuration:    Configuration  = Configuration()

        self._byProducts:   OrthogonalConnectorByProduct = cast(OrthogonalConnectorByProduct, None)

    @property
    def diagnosticInformation(self) -> DiagnosticInformation:
        return self._getDiagnosticInformation()

    # noinspection PyTypeChecker
    @classmethod
    def whichConnectorSide(cls, shape: UmlShapeGenre, anchorPosition: Position) -> Side:

        # shapeX, shapeY           = shape.GetPosition()
        # shapeWidth, shapeHeight  = shape.GetSize()

        umlPosition:   UmlPosition   = shape.position
        umlDimensions: UmlDimensions = shape.size

        shapeX: int = umlPosition.x
        shapeY: int = umlPosition.y
        shapeWidth: int = umlDimensions.width

        minX: int = shapeX
        maxX: int = shapeX + shapeWidth - ANCHOR_POINT_ADJUSTMENT
        minY: int = shapeY

        if anchorPosition.x == minX and anchorPosition.y >= minY:
            side: Side = Side.LEFT
        elif anchorPosition.x == maxX and anchorPosition.y >= minY:
            side = Side.RIGHT
        elif anchorPosition.x > minX and anchorPosition.y == minY:
            side = Side.TOP
        elif anchorPosition.x > minX and anchorPosition.y >= minY:
            side = Side.BOTTOM
        else:
            assert False, 'My algorithm has failed.  boo, hoo hoo'

        return side

    def runConnector(self, oglLink: UmlLinkGenre) -> bool:
        """

        Args:
            oglLink:

        Returns:  `True` the algorithm found a route, else `False`
        """

        sourceSide, destinationSide = self._determineAttachmentSide(umlLink=oglLink)

        sourceRect:      Rect = self._shapeToRect(oglLink.sourceShape)
        destinationRect: Rect = self._shapeToRect(oglLink.destinationShape)

        sourceConnectorPoint:      ConnectorPoint = ConnectorPoint(shape=sourceRect,      side=sourceSide,      distance=self._configuration.sourceEdgeDistance)
        destinationConnectorPoint: ConnectorPoint = ConnectorPoint(shape=destinationRect, side=destinationSide, distance=self._configuration.destinationEdgeDistance)

        options: OrthogonalConnectorOptions = OrthogonalConnectorOptions()
        options.pointA = sourceConnectorPoint
        options.pointB = destinationConnectorPoint

        options.shapeMargin        = self._configuration.shapeMargin
        options.globalBoundsMargin = self._configuration.globalBoundsMargin
        options.globalBounds       = self._configuration.globalBounds

        path: Points     = OrthogonalConnector.route(options=options)

        self._byProducts = OrthogonalConnector.byProduct

        self.logger.info(f'{path}')

        if len(path) == 0:      # noqa
            return False
        else:
            self._deleteTheOldLink(umlLink=oglLink)
            self._createOrthogonalLink(oldLink=oglLink, path=path)
            return True

    def _shapeToRect(self, umlShape: UmlShapeGenre) -> Rect:

        # shapeX, shapeY           = umlShape.GetPosition()
        # shapeWidth, shapeHeight  = umlShape.GetSize()
        umlPosition:   UmlPosition = umlShape.position
        umlDimensions: UmlDimensions = umlShape.size

        rect: Rect = Rect()

        rect.left   = umlPosition.x
        rect.top    = umlPosition.y
        rect.width  = umlDimensions.width
        rect.height = umlDimensions.height

        return rect

    def _determineAttachmentSide(self, umlLink: UmlLinkGenre) -> Tuple[Side, Side]:

        # sourceShape      = oglLink.sourceShape
        # destinationShape = oglLink.destinationShape
        # sourceAnchorPoint:      AnchorPoint = oglLink.sourceAnchor
        # destinationAnchorPoint: AnchorPoint = oglLink.destinationAnchor
        # sourcePosition:      Tuple[int, int] = sourceAnchorPoint.GetPosition()
        # destinationPosition: Tuple[int, int] = destinationAnchorPoint.GetPosition()
        # sourceSide:      Side = OrthogonalConnectorAdapter.whichConnectorSide(shape=sourceShape,      anchorPosition=Position(x=sourcePosition[0], y=sourcePosition[1]))
        # destinationSide: Side = OrthogonalConnectorAdapter.whichConnectorSide(shape=destinationShape, anchorPosition=Position(x=destinationPosition[0], y=destinationPosition[1]))

        sourceShape      = umlLink.sourceShape
        destinationShape = umlLink.destinationShape

        endPoints:           EndPositions = umlLink.endPositions
        sourcePosition:      UmlPosition  = endPoints.fromPosition
        destinationPosition: UmlPosition  = endPoints.toPosition

        self.logger.info(f'{sourcePosition=} {destinationPosition=}')

        sourceSide:      Side = OrthogonalConnectorAdapter.whichConnectorSide(
            shape=sourceShape,
            anchorPosition=Position(x=sourcePosition.x, y=sourcePosition.y)
        )
        destinationSide: Side = OrthogonalConnectorAdapter.whichConnectorSide(
            shape=destinationShape,
            anchorPosition=Position(x=destinationPosition.x, y=destinationPosition.y)
        )

        self.logger.info(f'{sourceSide=} {destinationSide=}')

        return sourceSide, destinationSide

    def _deleteTheOldLink(self, umlLink: UmlLinkGenre):

        self._extensionsFacade.deleteLink(umlLink=umlLink)

    def _createOrthogonalLink(self, oldLink: UmlLinkGenre, path: Points):

        linkType: LinkType = oldLink.modelLink.linkType
        if linkType == LinkType.INHERITANCE:
            umlInheritance:   UmlInheritance   = cast(UmlInheritance, oldLink)  # noqa
            sourceShape:      LinkableUmlShape = umlInheritance.subClass
            destinationShape: LinkableUmlShape = umlInheritance.baseClass
        else:
            sourceShape      = oldLink.sourceShape
            destinationShape = oldLink.destinationShape

        umlPositions: UmlPositions = self._toUmlPositions(path=path)

        linkInformation: LinkInformation = LinkInformation()
        linkInformation.linkType         = linkType
        linkInformation.path             = umlPositions
        linkInformation.sourceShape      = sourceShape
        linkInformation.destinationShape = destinationShape

        if linkType == LinkType.INTERFACE:
            linkInformation.interfaceName = InterfaceName(oldLink.modelLink.name)

        elif linkType == LinkType.ASSOCIATION or linkType == LinkType.COMPOSITION or linkType == LinkType.AGGREGATION:

            linkInformation.associationName        = AssociationName(oldLink.modelLink.name)
            linkInformation.sourceCardinality      = SourceCardinality(oldLink.modelLink.sourceCardinality)
            linkInformation.destinationCardinality = DestinationCardinality(oldLink.modelLink.destinationCardinality)

        self._extensionsFacade.createLink(linkInformation=linkInformation, callback=self._createLinkCallback)

    def _createLinkCallback(self, newLink: UmlLinkGenre):

        self._extensionsFacade.addShape(newLink)
        self._extensionsFacade.refreshFrame()

    def _toUmlPositions(self, path: Points) -> UmlPositions:

        oglPositions: UmlPositions = UmlPositions([])

        for pt in path:
            point:       Point       = cast(Point, pt)
            oglPosition: UmlPosition = UmlPosition(x=point.x, y=point.y)

            oglPositions.append(oglPosition)

        return oglPositions

    def _getDiagnosticInformation(self) -> DiagnosticInformation:
        """
        Don't leak OrthogonalConnector data types

        Returns:  Information that can be used to display why a routing connection failed
        """

        from pyorthogonalrouting.Point import Point
        from pyorthogonalrouting.Point import Points
        from pyorthogonalrouting.Rectangle import Rectangles

        from umlextensions.ExtensionsTypes import Point as DiagnosticPoint
        from umlextensions.ExtensionsTypes import Points as DiagnosticPoints
        from umlextensions.ExtensionsTypes import Rectangle as DiagnosticRectangle
        from umlextensions.ExtensionsTypes import Rectangles as DiagnosticRectangles
        from umlextensions.ExtensionsTypes import IntegerList

        def toDiagnosticPoints(refPoints:  Points) -> DiagnosticPoints:

            diagnosticPts: DiagnosticPoints = DiagnosticPoints([])

            for pt in refPoints:
                point: Point = cast(Point, pt)
                self.logger.info(f'{point=}')

                diagnosticPoint: DiagnosticPoint = DiagnosticPoint(x=point.x, y=point.y)
                diagnosticPts.append(diagnosticPoint)

            return diagnosticPts

        def toDiagnosticRectangle(rectangle: Rectangle) -> DiagnosticRectangle:

            return DiagnosticRectangle(
                left=rectangle.left,
                top=rectangle.top,
                width=rectangle.width,
                height=rectangle.height
            )

        def toIntegerList(integers: Integers) -> IntegerList:

            integerList: IntegerList = IntegerList([])
            for integer in integers:
                integerList.append(integer)

            return integerList

        def toDiagnosticRectangles(rectangles: Rectangles) -> DiagnosticRectangles:

            diagnosticRectangles: DiagnosticRectangles = DiagnosticRectangles([])
            for r in rectangles:
                diagnosticRectangle: DiagnosticRectangle = toDiagnosticRectangle(r)
                diagnosticRectangles.append(diagnosticRectangle)

            return diagnosticRectangles

        diagnosticInformation: DiagnosticInformation = DiagnosticInformation(
            spots=toDiagnosticPoints(refPoints=self._byProducts.spots),
            horizontalRulers=toIntegerList(self._byProducts.hRulers),
            verticalRulers=toIntegerList(self._byProducts.vRulers),
            diagramBounds=toDiagnosticRectangle(self._byProducts.diagramBounds),
            routeGrid=toDiagnosticRectangles(self._byProducts.grid)
        )

        return diagnosticInformation




from typing import Callable
from typing import cast

from logging import Logger
from logging import getLogger

from wx import BoxSizer
from wx import CommandEvent
from wx import EVT_SLIDER
from wx import EXPAND
from wx import ID_ANY
from wx import LEFT
from wx import Panel
from wx import SL_AUTOTICKS
from wx import SL_HORIZONTAL
from wx import SL_LABELS
from wx import Size
from wx import DefaultSize
from wx import Slider
from wx import VERTICAL

from wx.lib.sized_controls import SizedPanel
from wx.lib.sized_controls import SizedStaticBox

ValueChangedHandler = Callable[[CommandEvent], None]


class LabelledSlider:
    """
    The impetus of this component was to work around a bug that I described
    at stack overflow https://stackoverflow.com/questions/78414996/wxpython-slider-incorrectly-displays-with-sized-panels/78425220#78425220
    and that Richard Townsend bull dogged on to get me a workaround.

    (https://discuss.wxpython.org/t/wxpython-slider-incorrectly-displays-with-sized-panels/36915/8)

    Since bare sliders always need labels, voila, here is the wrapper component

    """
    def __init__(self, sizedPanel: SizedPanel, label: str, value: int, minValue: int, maxValue, size: Size = DefaultSize, toolTip: str = ''):
        """

        Args:
            sizedPanel:     Parent Sized Panel
            label:          The label to display
            value:          The value to set the control to
            minValue:       The slider minimal value
            maxValue:       The slider maximum value
            size:           The Size of the component
            toolTip:        A tool tip to display on a mouse over
        """

        self.logger: Logger = getLogger(__name__)

        self._slider:   Slider              = cast(Slider, None)
        self._callback: ValueChangedHandler = cast(ValueChangedHandler, None)

        self._layoutComponent(sizedPanel, label, value=value, minValue=minValue, maxValue=maxValue, size=size, toolTip=toolTip)

        sizedPanel.Bind(EVT_SLIDER, handler=self._onSliderChanged, source=self._slider)

    def _valueChangedHandler(self, handler: ValueChangedHandler):
        self._callback = handler

    # noinspection PyTypeChecker
    valueChangedHandler = property(fget=None, fset=_valueChangedHandler, doc='This method called when slider value changes')

    def _layoutComponent(self, parentPanel: SizedPanel, label: str, value: int, minValue: int, maxValue, size: Size, toolTip: str):

        sizedStaticBox: SizedStaticBox = SizedStaticBox(parent=parentPanel, label=label)
        sizedStaticBox.SetSizerType('vertical')
        sizedStaticBox.SetSizerProps(expand=True, proportion=1)

        sliderPanel: Panel  = Panel(sizedStaticBox)

        sliderStyle: int    = SL_HORIZONTAL | SL_AUTOTICKS | SL_LABELS
        slider:      Slider = Slider(sliderPanel, id=ID_ANY, value=value, minValue=minValue, maxValue=maxValue, size=size, style=sliderStyle)

        sliderPanelSizer: BoxSizer = BoxSizer(VERTICAL)
        # Add (self, window, proportion=0, flag=0, border=0, userData=None)
        sliderPanelSizer.Add(slider, 1, EXPAND | LEFT, 5)
        sliderPanel.SetSizer(sliderPanelSizer)

        sliderPanel.Layout()

        slider.SetToolTip(toolTip)

        self._slider = slider

    def _onSliderChanged(self, event: CommandEvent):

        assert self._callback is not None, 'Developer error.  You want to set the valueChangedHandler'
        self._callback(event)



#
#  These types are visible to the extension adapter and the
#  implementor of the adapter
#
from typing import List
from typing import NewType
from typing import Tuple
from typing import cast
from typing import Callable

from dataclasses import dataclass
from dataclasses import field

from umlshapes.ShapeTypes import UmlLinkGenre
from umlshapes.types.UmlPosition import UmlPositions
from wx import ClientDC

from umlmodel.enumerations.LinkType import LinkType

from umlshapes.ShapeTypes import LinkableUmlShape
from umlshapes.frames.UmlFrame import UmlFrame

from umlshapes.ShapeTypes import UmlShapes
from umlshapes.ShapeTypes import umlShapesFactory

from umlio.IOTypes import UmlDocumentType

InterfaceName          = NewType('InterfaceName', str)
AssociationName        = NewType('AssociationName', str)
SourceCardinality      = NewType('SourceCardinality', str)
DestinationCardinality = NewType('DestinationCardinality', str)

@dataclass
class FrameSize:
    """
    The strategy is to provide minimal information to the extensions
    we do not want them to not abuse it.
    """
    width:  int = -1
    height: int = -1


def createFrameSizeFactory() -> FrameSize:
    """
    Factory method to create  the OglClasses data structure;

    Returns:  A new data structure
    """
    return FrameSize()


@dataclass
class FrameInformation:
    """
    The document title is the name of the frame
    """
    umlFrame:           UmlFrame   = cast(UmlFrame, None)
    frameActive:        bool       = False
    selectedUmlShapes:  UmlShapes  = field(default_factory=umlShapesFactory)
    frameSize:          FrameSize  = field(default_factory=createFrameSizeFactory)
    clientDC:           ClientDC   = cast(ClientDC, None)
    diagramTitle:       str        = ''
    diagramType:        UmlDocumentType = UmlDocumentType.NOT_SET


FrameInformationCallback  = Callable[[FrameInformation], None]
FrameSizeCallback         = Callable[[FrameSize], None]
SelectedUmlShapesCallback = Callable[[UmlShapes], None]

NO_INTEGER: int = cast(int, None)


@dataclass
class ShapeBoundaries:
    minX: int = NO_INTEGER
    minY: int = NO_INTEGER
    maxX: int = NO_INTEGER
    maxY: int = NO_INTEGER


ObjectBoundaryCallback = Callable[[ShapeBoundaries], None]

@dataclass(eq=True)
class Point:
    """
    A point in space.
    """
    x: int = NO_INTEGER
    y: int = NO_INTEGER

    def toTuple(self) -> Tuple[int, int]:
        return self.x, self.y

    def __hash__(self) -> int:
        return hash((self.x, self.y))


IntegerList = NewType('IntegerList', List[int])
Points      = NewType('Points',      List[Point])

@dataclass
class Rectangle:
    left:   int = 0
    top:    int = 0
    width:  int = 0
    height: int = 0


Rectangles = NewType('Rectangles', List[Rectangle])

@dataclass
class DiagnosticInformation:

    horizontalRulers: IntegerList
    verticalRulers:   IntegerList
    diagramBounds:    Rectangle
    spots:            Points
    routeGrid:        Rectangles

def umlPositionsFactory() -> UmlPositions:
    return UmlPositions([])

@dataclass
class LinkInformation:
    """
    The field interfaceName is only valid when linkType is PyutLinkType.INTERFACE
    The fields
        associationName
        sourceCardinality
        destinationCardinality
    are valid only when linkType is  one of
        PyutLinkType.ASSOCIATION
        PyutLinkType.COMPOSITION
        PyutLinkType.AGGREGATION
    """
    linkType:               LinkType                = cast(LinkType, None)
    path:                   UmlPositions            = field(default_factory=umlPositionsFactory)
    sourceShape:            LinkableUmlShape        = cast(LinkableUmlShape, None)
    destinationShape:       LinkableUmlShape        = cast(LinkableUmlShape, None)
    interfaceName:          InterfaceName           = cast(InterfaceName, None)
    associationName:        AssociationName         = cast(AssociationName, None)
    sourceCardinality:      SourceCardinality       = cast(SourceCardinality, None)
    destinationCardinality: DestinationCardinality  = cast(DestinationCardinality, None)


CreatedLinkCallback = Callable[[UmlLinkGenre], None]




from abc import ABC
from abc import abstractmethod

from umlshapes.ShapeTypes import UmlLinkGenre
from umlshapes.ShapeTypes import UmlShapeGenre

from umlshapes.pubsubengine.IUmlPubSubEngine import IUmlPubSubEngine

from umlextensions.ExtensionsPubSub import ExtensionsPubSub

from umlextensions.ExtensionsTypes import CreatedLinkCallback
from umlextensions.ExtensionsTypes import FrameInformation
from umlextensions.ExtensionsTypes import FrameInformationCallback
from umlextensions.ExtensionsTypes import IntegerList
from umlextensions.ExtensionsTypes import LinkInformation
from umlextensions.ExtensionsTypes import ObjectBoundaryCallback
from umlextensions.ExtensionsTypes import Points
from umlextensions.ExtensionsTypes import Rectangle
from umlextensions.ExtensionsTypes import Rectangles
from umlextensions.ExtensionsTypes import SelectedUmlShapesCallback


class IExtensionsFacade(ABC):
    """
    This facade simplifies communication to the UML diagrammer.  This interface serves as a front-facing interface
    that masks the complexity of the UML Diagrammer
    """
    def __init__(self):
        self._extensionsPubSub: ExtensionsPubSub = ExtensionsPubSub()

    @property
    def extensionsPubSub(self) -> ExtensionsPubSub:
        return self._extensionsPubSub

    @property
    @abstractmethod
    def umlPubSub(self) -> IUmlPubSubEngine:
        pass

    @umlPubSub.setter
    @abstractmethod
    def umlPubSub(self, umlPubSub: IUmlPubSubEngine):
        pass

    @abstractmethod
    def requestCurrentFrameInformation(self, callback: FrameInformationCallback) -> FrameInformation:
        pass

    @abstractmethod
    def extensionModifiedProject(self):
        pass

    @abstractmethod
    def selectUmlShapes(self):
        pass

    @abstractmethod
    def getSelectedUmlShapes(self, callback: SelectedUmlShapesCallback):
        pass

    @abstractmethod
    def refreshFrame(self):
        pass

    @abstractmethod
    def addShape(self, umlShape: UmlShapeGenre | UmlLinkGenre):
        pass

    @abstractmethod
    def wiggleShapes(self):
        pass

    @abstractmethod
    def getShapeBoundaries(self, callback: ObjectBoundaryCallback):
        """
        Request the boundaries around all the UML objects
        on the current frame

        Args:
            callback:  The callback that receives the boundaries
        """
        pass

    @abstractmethod
    def deleteLink(self, umlLink: UmlLinkGenre):
        pass

    @abstractmethod
    def createLink(self, linkInformation: LinkInformation, callback: CreatedLinkCallback):
        pass

    @abstractmethod
    def showOrthogonalRoutingPoints(self, show: bool, spots: Points):
        pass

    @abstractmethod
    def showRulers(self, show: bool, horizontalRulers: IntegerList, verticalRulers: IntegerList, diagramBounds: Rectangle):
        pass

    @abstractmethod
    def showRouteGrid(self, show: bool, routeGrid: Rectangles):
        pass




from typing import cast

from pathlib import Path

from wx import BITMAP_TYPE_PNG
from wx import Bitmap
from wx import ClientDC
from wx import Image
from wx import MemoryDC
from wx import NullBitmap
from wx import Window

from wx import BitmapType

from umlextensions.ExtensionsTypes import FrameInformation

NO_PARENT_WINDOW: Window = cast(Window, None)


def createScreenImageFile(frameInformation: FrameInformation, imagePath: Path, imageType: BitmapType = BITMAP_TYPE_PNG) -> bool:
    """
    Create a screen image file
    Args:
        frameInformation:   Plugin frame information
        imagePath:          Where to write the image file to
        imageType:          Defaults to png

    Returns: 'True' for a successful creation else 'False'

    """

    context:   ClientDC   = frameInformation.clientDC
    memory:    MemoryDC   = MemoryDC()

    x: int = frameInformation.frameSize.width
    y: int = frameInformation.frameSize.height
    emptyBitmap: Bitmap = Bitmap(x, y, -1)

    memory.SelectObject(emptyBitmap)
    memory.Blit(source=context, xsrc=0, height=y, xdest=0, ydest=0, ysrc=0, width=x)
    memory.SelectObject(NullBitmap)

    img: Image = emptyBitmap.ConvertToImage()

    status: bool = img.SaveFile(str(imagePath), imageType)

    return status




from typing import cast

from logging import Logger
from logging import getLogger

from os import linesep as osLineSep

from wx import OK
from wx import ICON_ERROR
from wx import MessageDialog

from umlmodel.enumerations.LinkType import LinkType

from umlshapes.ShapeTypes import UmlLinkGenre

from umlextensions.Common import NO_PARENT_WINDOW

from umlextensions.IExtensionsFacade import IExtensionsFacade
from umlextensions.extensiontypes.ExtensionDataTypes import Author
from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionName
from umlextensions.extensiontypes.ExtensionDataTypes import Version

from umlextensions.tools.BaseToolExtension import BaseToolExtension
from umlextensions.tools.orthogonalrouting.DlgDiagnoseLayout import DlgDiagnoseLayout
from umlextensions.tools.orthogonalrouting.DlgOrthogonalRoutingConfiguration import DlgOrthogonalRoutingConfiguration
from umlextensions.tools.orthogonalrouting.OrthogonalConnectorAdapter import OrthogonalConnectorAdapter


class ToolOrthogonalRouting(BaseToolExtension):

    def __init__(self, extensionsFacade: IExtensionsFacade):

        super().__init__(extensionsFacade=extensionsFacade)
        self.logger: Logger = getLogger(__name__)

        self._name      = ExtensionName('Orthogonal Links')
        self._author    = Author('Humberto A. Sanchez II')
        self._version   = Version('2.0')

    def setOptions(self) -> bool:
        """
        Prepare for the tool action.
        This can be used to ask some questions to the user.

        Returns: If False, the import should be cancelled.  'True' to proceed
        """
        with DlgOrthogonalRoutingConfiguration(NO_PARENT_WINDOW, extensionsFacade=self._extensionsFacade) as dlg:
            if dlg.ShowModal() == OK:
                return True
            else:
                self.logger.warning(f'Cancelled')
                return False

    def doAction(self):

        self.logger.info(f'Begin Orthogonal Routing')
        self._extensionsFacade.getSelectedUmlShapes(callback=self._doAction)
        self.logger.info('End Orthogonal Routing')

    def _doAction(self, selectedUmlShapes):
        self.logger.info(f'_doAction')

        adapter: OrthogonalConnectorAdapter = OrthogonalConnectorAdapter(extensionsFacade=self._extensionsFacade)

        umlLink: UmlLinkGenre = cast(UmlLinkGenre, None)

        for potentialLink in selectedUmlShapes:
            if isinstance(potentialLink, UmlLinkGenre):
                try:
                    umlLink = cast(UmlLinkGenre, potentialLink)
                    success: bool = adapter.runConnector(oglLink=umlLink)
                    if success is False:
                        message: str           = self._composeGoodErrorMessage(umlLink)
                        booBoo:  MessageDialog = MessageDialog(parent=None, message=message, caption='No orthogonal route', style=OK | ICON_ERROR)
                        booBoo.ShowModal()
                        if self._preferences.diagnoseOrthogonalRouter is True:

                            dlg: DlgDiagnoseLayout    = DlgDiagnoseLayout(parent=None)
                            dlg.extensionsFacade      = self._extensionsFacade
                            dlg.diagnosticInformation = adapter.diagnosticInformation
                            dlg.Show(True)
                        break

                except (AttributeError, TypeError) as e:
                    self.logger.error(f'{e} - {umlLink=}')

    def _composeGoodErrorMessage(self, oglLink: UmlLinkGenre) -> str:

        linkType: LinkType = oglLink.modelLink.linkType
        message: str = (
            f'Could not find an orthogonal route for link: {linkType}{osLineSep}'
            f'from {osLineSep}'
            f'{oglLink.sourceShape} {osLineSep}' 
            f'to{osLineSep}'
            f'{oglLink.destinationShape}'
        )

        return message




from typing import cast

from logging import Logger
from logging import getLogger

from umlextensions.IExtensionsFacade import IExtensionsFacade

from umlextensions.extensiontypes.ExtensionDataTypes import Author
from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionDescription
from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionName
from umlextensions.extensiontypes.ExtensionDataTypes import FileSuffix
from umlextensions.extensiontypes.ExtensionDataTypes import FormatName
from umlextensions.extensiontypes.ExtensionDataTypes import Version
from umlextensions.extensiontypes.SingleFileRequestResponse import SingleFileRequestResponse

from umlextensions.output.OutputFormat import OutputFormat
from umlextensions.output.BaseOutputExtension import BaseOutputExtension
from umlextensions.output.gml.GMLExporter import GMLExporter

FORMAT_NAME:           FormatName           = FormatName('GML')
FILE_SUFFIX:           FileSuffix           = FileSuffix('gml')
EXTENSION_DESCRIPTION: ExtensionDescription = ExtensionDescription('Graph Modeling Language - Portable Format for Graphs')

class OutputGML(BaseOutputExtension):

    def __init__(self, extensionsFacade: IExtensionsFacade):
        
        super().__init__(extensionsFacade=extensionsFacade)
        self.logger: Logger = getLogger(__name__)

        self._name    = ExtensionName('UML to GML')
        self._author  = Author('Humberto A. Sanchez II')
        self._version = Version('4.0')

        self._outputFormat  = OutputFormat(formatName=FORMAT_NAME, fileSuffix=FILE_SUFFIX, description=EXTENSION_DESCRIPTION)

        self._exportResponse: SingleFileRequestResponse = cast(SingleFileRequestResponse, None)

    def setExportOptions(self) -> bool:

        defaultFileName: str = self._preferences.defaultGMLFilename

        self._exportResponse = self.askForFileToExport(defaultFileName=defaultFileName)

        if self._exportResponse.cancelled is True:
            return False
        else:
            return True

    def write(self) -> bool:

        gmlExporter: GMLExporter = GMLExporter()

        selectedUmlShapes = self._frameInformation.selectedUmlShapes
        gmlExporter.translate(umlShapes=selectedUmlShapes)

        gmlExporter.write(self._exportResponse.fileName)

        return True
